#!/usr/bin/env python

"""
Given already mapped fusions using the reads file (format:
gene1 gene2 position1 strand1 position2 strand2 read_name)
Use the original BAM to plot the ends of the reads as BED file to be presented
by the genome browser.
Color code as specified in the parametrs
"""

import sys
import argparse
import csv
from collections import defaultdict
import random

from Bio.Seq import Seq
import pysam
from Bio import SeqIO
import RILseq

def process_command_line(argv):
    """
    Return a 2-tuple: (settings object, args list).
    `argv` is a list of arguments, or `None` for ``sys.argv[1:]``.
    """
    if argv is None:
        argv = sys.argv[1:]

    # initialize the parser object:
    parser = argparse.ArgumentParser(
        description='Generate BED graph of the reads.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        'genome',
        help='genome fasta file.')
    parser.add_argument(
        'list_reads',
        help='File with list of reads and their fused positions.')
    parser.add_argument(
        'track_name',
        help='Name of track')
    parser.add_argument(
        'track_desc',
        help='Description of the track')
    parser.add_argument(
        'bamfiles', action='append', nargs='+',
        help='The original bam file (or several files) with the full reads.')
    parser.add_argument(
        '-r', '--reverse', default=False, action='store_true',
        help='The original bam file is the reverse complement of the RNA.')
    parser.add_argument(
        '-s', '--summary',
        help='Print only reads that are found to be significant in this summary file.')
    parser.add_argument(
        '-e', '--gene_name',
        help='Print reads involve only this gene (EcoCyc ID), '
        'applies only with -s')
    parser.add_argument(
        '--rand_score', default=False, action='store_true',
        help='Set a random score (0-1000) for each read, this will allow to '
        'present some of the reads in UCSC genome browser.')
    parser.add_argument(
        '--pos_first', default='255,0,0',
        help='Color of first part, positive strand.')
    parser.add_argument(
        '--pos_second', default='51,102,255',
        help='Color of second part, positive strand.')
    parser.add_argument(
        '--rev_first', default='255,0,0',
        help='Color of first part, reverse strand.')
    parser.add_argument(
        '--rev_second', default='51,102,255',
        help='Color of second part, reverse strand.')
    parser.add_argument(
        '--BC_chrlist', default='COLI-K12,chr',
        help='A comma separated dictionary of chromosome names from the BioCyc'
        ' names to the bam file names. The names in the bam file should be '
        ' the same as the UCSC genome browser (they will be printed).')
    settings = parser.parse_args(argv)
    return settings


def get_reads_seqs(bamfile, rnames, rev=False):
    """
    Return the sequences of all the reads from the bam file
    Arguments:
    - `bamfile`: The pysam file
    - `rnames`: reads names (a list of the 5p read id)
    """
    r1_seqs = {}
    r2_seqs = {}
    rqns = set()
    reads = defaultdict(list)
    for read in bamfile.fetch(until_eof=True):
        rqns.add(read.query_name)
        reads[read.query_name].append(read)  # read reads
    for rn in set(rnames) & rqns:  # iterate through the reads which are also in the bam file
        for read in reads[rn]:
            if read.is_read1==rev:  # read is first and reverse flag (Livny) this actually tests for 3p
                outseq = Seq(read.query_sequence)
                if not read.is_reverse:  # if read is not reversed, reverse_complement the read and store in r1_seqs
                    outseq = outseq.reverse_complement()
                r1_seqs[read.query_name] = str(outseq)
            else:  # read is second and reverse flag (Livny). this actually tests for 5p
                outseq = Seq(read.query_sequence)
                if read.is_reverse:  # if read is reversed, reverse_complement the read and store in r2_seqs
                    outseq = outseq.reverse_complement()
                r2_seqs[read.query_name] = str(outseq)
    # r1_seqs is the 3' end of the second fused RNA, r2_seqs is the 5' of the first fused RNA
    return r1_seqs, r2_seqs


def extend_alignment(rseq, pos5p, pos3p, is_read1, strand, genome, mismatch=1):
    """
    Align the rseq to the genome in the specified position. Return the last
    position of the read mapped to the genome.
    Use local alignment
    Arguments:
    - `rseq`: Read sequence
    - `pos5p`: the 5' position, exact if read 2 or as limit if read 1
    - `pos3p`: the 3' position, exact if read 1 or as limit if read 2
    - `is_read1`: This read is read 1
    - `strand`: mapping strand
    - `genome`: The genome Seq object
    - `mismatch`: allowed mismatches

    side_5p_len = extend_alignment(
        s1+overlap+s2, read_5ps[rname][0], 0, False, read_5ps[rname][1],
        genome[read_5ps[rname][2]])
    side_3p_len = extend_alignment(
        s1+overlap+s2, 0, read_3ps[rname][0], True, read_3ps[rname][1],
        genome[read_3ps[rname][2]])

    """
    rcnt = {'A':'T', 'C':'G', 'G':'C', 'T':'A'}
    glen = len(genome)
    if is_read1:
        # Start from the last position and move on to the 5' end
        if strand == '-':
            ipos = 0
            for _ in range(mismatch+1):
                try:
                    while rcnt[genome[(pos3p+ipos)%glen]] == rseq[-(ipos+1)]:
                        ipos += 1
                except IndexError:
                    return ipos - 1
                ipos += 1
            return ipos
        else:
            ipos = 0
            for _ in range(mismatch+1):
                try:
                    while genome[(pos3p-ipos)%glen] == rseq[-(ipos+1)]:
                        ipos += 1
                except IndexError:
                    return ipos-1
                ipos += 1
            return ipos
    else:
        if strand == '-':
            ipos = 0
            for _ in range(mismatch+1):
                try:
                    while rcnt[genome[(pos5p-ipos)%glen]] == rseq[ipos]:
                        ipos += 1
                except IndexError:
                    return ipos - 1
                ipos += 1
            return ipos 
        else:
            ipos = 0
            for _ in range(mismatch+1):
                try:
                    while genome[(pos5p+ipos)%glen] == rseq[ipos]:
                        ipos += 1
                except IndexError:
                    return ipos - 1
                ipos += 1
            return ipos 
        

def find_overlap(s1, s2):
    """
    Find overlaps between two reads. Assume they are both in the same
    orientation (r1 is revcomp)
    Return 3 sequences: s1, overlap, s2
    e.g: find_overlap("ACTGTGTGTGTGCC", "GTGTGCCCCCCCCCCC") - ('ACTGTGT', 'GTGTGCC', 'CCCCCCCCC')
    Arguments:
    - `s1`: first sequence, this is mate 2 actually in our experiments
    - `s2`: last sequence, mate 1
    """
    for i in range(min(len(s1), len(s2)))[::-1]:
        if s1[-i:]==s2[:i]:
            return s1[:-i], s1[-i:], s2[i:]
    return s1, '', s2


def main(argv=None):
    settings = process_command_line(argv)
    # Read the read names and positions
    read_5ps = {}
    read_3ps = {}
    read_genes = {}
    genome = {}
    gsize = {}
    for sr in SeqIO.parse(settings.genome, 'fasta'):
        genome[sr.id] = sr.seq
        gsize[sr.id] = len(sr.seq)  # genome size dictionary - {chr:size}
    if len(settings.BC_chrlist) >= 2:
        chr_dict = dict(zip(
                settings.BC_chrlist.split(',')[0::2],
                settings.BC_chrlist.split(',')[1::2]))
    #  create dictionary of {'COLI-K12' : 'chr'}
    else:
        chr_dict = {}
    if settings.summary:  # only reads from the significant interactions in -s param, also can enter a specific gene.
        sig_reads = RILseq.read_significant_reads(
            settings.summary, chr_dict, gname=settings.gene_name)

    for line in csv.reader(open(settings.list_reads), delimiter='\t'):
        # skip single
        if len(line) > 7 and line[7]=="single":
            continue
        if settings.summary:
            if (int(line[4])-1, line[5], line[3]) not in\
                    sig_reads[(int(line[1])-1, line[2], line[0])]:
                # skip if (coord_2, strand_2, chrom_2) not in the (coord_2, strand_2, chrom_2) of the significant reads.
                continue
        read_5ps[line[6]] = [int(line[1])-1, line[2], line[0]]  # {read_id : [coord_1(0-based), strand_1, chrom_1]}
        read_3ps[line[6]] = [int(line[4])-1, line[5], line[3]]  # {read_id : [coord_2(0-based), strand_2, chrom_2]}
#        read_genes[line[6]] = [line[0], line[1]]
    # Read the bam files and return the long sequences
    r1_seqs = {}
    r2_seqs = {}
    for bamfile in list(RILseq.flat_list(settings.bamfiles)):  # flat multiple lists into one list.
        r1s, r2s = get_reads_seqs(
            pysam.AlignmentFile(bamfile, 'rb'), read_5ps.keys(), rev=settings.reverse)
        r1_seqs.update(r1s)
        r2_seqs.update(r2s)
    # For each read find the overlap, if exists and find the fusion point
    outer = csv.writer(sys.stdout, delimiter='\t')
    print ('track name="%s" description="%s" visibility=4 itemRgb="On" useScore=0'%(
        settings.track_name, settings.track_desc))
    # Because I'm lazy, the code is written so r1 is the 3' end of the fragment
    for rname in set(r2_seqs.keys()):
        if rname in r1_seqs:
            r2seq = r2_seqs[rname]
            r1seq = r1_seqs[rname]
        else:  # single-end
            r2seq = r2_seqs[rname]
            r1seq = ''
        # r2seq, r1seq are the sequences from the bam files for paired end
        s1, overlap, s2 = find_overlap(r2seq, r1seq)
        # print here
        print(s1+overlap+s2)
        print(read_5ps[rname][0])
        print(read_3ps[rname][0])
        side_5p_len = extend_alignment(
            s1+overlap+s2, read_5ps[rname][0], 0, False, read_5ps[rname][1],
            genome[read_5ps[rname][2]])
        side_3p_len = extend_alignment(
            s1+overlap+s2, 0, read_3ps[rname][0], True, read_3ps[rname][1],
            genome[read_3ps[rname][2]])
        # Write each of the sides to the output file
        score=0
        if settings.rand_score:
            score=random.randint(0, 1000)
        if read_5ps[rname][1] == '+':
            gfrom = max(0,  read_5ps[rname][0])
            gto = min(gsize[read_5ps[rname][2]], read_5ps[rname][0]+side_5p_len)
            outer.writerow([
                    read_5ps[rname][2], gfrom, gto, "%s_5p"%rname, score, '+',
                    gfrom, gto, settings.pos_first])
        elif read_5ps[rname][1] == '-':
            gfrom = max(0, read_5ps[rname][0]-side_5p_len+1)
            gto = min(gsize[read_5ps[rname][2]], read_5ps[rname][0]+1)
            outer.writerow([
                    read_5ps[rname][2], gfrom, gto, "%s_5p"%rname, score, '-',
                    gfrom, gto, settings.rev_first])
        if read_3ps[rname][1] == '+':
            gfrom = max(0, read_3ps[rname][0]-side_3p_len+1)
            gto = min(gsize[read_3ps[rname][2]], read_3ps[rname][0]+1)
            outer.writerow([
                    read_3ps[rname][2], gfrom, gto,"%s_3p"%rname, score, '+',
                    gfrom, gto, settings.pos_second])
        elif read_3ps[rname][1] == '-':
            gfrom = max(0, read_3ps[rname][0])
            gto = min(gsize[read_3ps[rname][2]], read_3ps[rname][0]+side_3p_len)
            outer.writerow([
                    read_3ps[rname][2], gfrom, gto, "%s_3p"%rname, score, '-',
                    gfrom, gto, settings.rev_second])
    return 0        # success

if __name__ == '__main__':
    status = main()
    sys.exit(status)



#!/usr/bin/env python

"""
Read the list of chimeric interactions and generate a file that can be read
by circos.
"""

import sys
import argparse
from collections import defaultdict
from math import log
import gzip

import RILseq

def process_command_line(argv):
    """
    Return a 2-tuple: (settings object, args list).
    `argv` is a list of arguments, or `None` for ``sys.argv[1:]``.
    """
    if argv is None:
        argv = sys.argv[1:]

    # initialize the parser object, replace the description
    parser = argparse.ArgumentParser(
        description='Generate circos data file.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        'reads_in',
        help='An output file of map_chimeric_fragments.py with the chimeric'
        ' fragments.')
    parser.add_argument(
        '-s', '--summary',
        help='Plot only significant interactions that appear in the summary file.')
    parser.add_argument(
        '-r', '--region', type=int, default=200,
        help='Split the genome to windows of this size.')
    parser.add_argument(
        '--sRNAs', default=False, action='store_true',
        help='Color the lines going to or coming from sRNAs in orange. Must be'
        ' used with ec_dir.')
    parser.add_argument(
        '--known',
        help='Use this file to color the known interactions in red.'
        ' Must give --refseq_dir as well.')
    parser.add_argument(
        '--refseq_dir', default='/home/users/assafp/EC/',
        help='RefSeq dir of organism to get the gene description from.'
        ' Should be given if --known is given.')
    parser.add_argument(
        '--ec_dir', #default='/home/users/assafp/Database/EcoCyc/19.0/data',
        help='EcoCyc data dir, used to map the regions to genes. If not'
        ' given only the regions will be reported')
    parser.add_argument(
        '-c', '--chrn', default='chr',
        help='Name of chromosome to plot.')
    parser.add_argument(
        '-p', '--print_chr', default='ecmain',
        help='Name of chromosome in circos.')
    parser.add_argument(
        '-m', '--min_interactions', type=int, default=100,
        help='Minimum number of interactions between two regions to plot.')
    parser.add_argument(
        '--EC_chrlist', default='COLI-K12,chr',
        help='A comma separated dictionary of chromosome names from the EcoCyc'
        ' names to the bam file names. The names in the bam file should be '
        ' the same as the UCSC genome browser (they will be printed).')
    settings = parser.parse_args(argv)

    return settings

def get_coords(tfile):
    """
    Return the coords of genes in the file
    Arguments:
    - `tfile`: ptt.gz or rnt.gz file
    """
    coors = {}
    strand = {}
    tin = gzip.open(tfile)
    for line in tin:
        spl = line.strip().split()
        try:
            t1, t2 = spl[0].split('..')
            coors[spl[4]] = (int(t1)-1, int(t2))
            strand[spl[4]] = spl[1]
        except:
            pass
    return coors, strand

def main(argv=None):
    settings = process_command_line(argv)
    if len(settings.EC_chrlist) >= 2:
        chr_dict = dict(zip(
                settings.EC_chrlist.split(',')[0::2],
                settings.EC_chrlist.split(',')[1::2]))
    else:
        chr_dict = {}
    region_interactions, _, _, _=\
        RILseq.read_reads_table(open(settings.reads_in), settings.region)
    both_strs = defaultdict(lambda: defaultdict(int))
    if settings.summary:
        sig_reads = RILseq.read_significant_reads(
            settings.summary, chr_dict)
    if settings.known:
        known_reads = defaultdict(list)
        ptt_c, ptt_str = get_coords("%s.ptt.gz"%settings.refseq_dir)
        rnt_c, rnt_str = get_coords("%s.rnt.gz"%settings.refseq_dir)
        for line in open(settings.known):
            spl = line.strip().split()
            try:
                scoor = rnt_c[spl[0]]
                rcoor = ptt_c[spl[1]]
            except KeyError:
                pass
            else:
                for i in range(scoor[0], scoor[1]):
                    for j in range(rcoor[0], rcoor[1]):
                        known_reads[i].append(j)
                        known_reads[j].append(i)

    for reg1 in region_interactions:
        if reg1[2] != settings.chrn:
            continue
        for reg2 in region_interactions[reg1]:
            if reg2[2] != settings.chrn:
                continue
            if settings.summary:
                nsigs = 0
                for r1, r2, in region_interactions[reg1][reg2]:
                    nsigs += int((r2, reg2[1], reg2[2]) in \
                        sig_reads[(r1, reg1[1], reg1[2])])
            else:
                nsigs = len(region_interactions[reg1][reg2])
            both_strs[reg1[0]][reg2[0]] += nsigs
    if settings.sRNAs:
        from RILseq.ecocyc_parser import read_genes_data
        uid_pos, uid_names, uid_tudata, sRNAs_list, other_RNAs_list, rRNAs = \
            read_genes_data(settings.ec_dir)
        sposs = set()
        for g in sRNAs_list:
            for i in range(uid_pos[g][1], uid_pos[g][2]):
                sposs.add(i)
    for r1 in both_strs:
        for r2 in both_strs[r1]:
            if both_strs[r1][r2] > settings.min_interactions:
                color = 'thickness=%dp'%max(
                    int(log(both_strs[r1][r2])/log(10)),1)
                if settings.sRNAs:
                    rset = set([i for i in range(r1, r1+settings.region)])
                    rset |= set([i for i in range(r2, r2+settings.region)])
                    if rset & sposs:
                        color = 'color=orange'
                if settings.known:
                    for k in set(range(r1, r1+settings.region)) & set(known_reads.keys()):
                        if set(range(r2, r2+settings.region)) & set(known_reads[k]):
                            color = 'color=red'
                            
                sys.stdout.write('%s %d %d %s %d %d %s\n'%(
                        settings.print_chr, r1+1, r1+settings.region,
                        settings.print_chr, r2+1, r2+settings.region,
                        color))

    return 0        # success

if __name__ == '__main__':
    status = main()
    sys.exit(status)



#!/usr/bin/env python

"""
After a library is mapped to the genome (using map_single_fragments.py or any
other mapper), the bam file is screened for reads that weren't mapped to the
genome or weren't concise and try to map wach of the ends to a different
location. This script report the reads that are chimeric in a table of the
format:
chr1   position1    strand1    chr2   position2    strand2    read_name     read_type

where the position1 is the first position of the first read and position2 is
the last position of read2.
The input is a list of bam files, the output is always one list. The list can
be separated afterwards according to read names.
"""

import sys
import argparse
import pysam
import os
import errno
import pkg_resources

import RILseq

def process_command_line(argv):
    """
    Return a 2-tuple: (settings object, args list).
    `argv` is a list of arguments, or `None` for ``sys.argv[1:]``.
    """
    if argv is None:
        argv = sys.argv[1:]

    # initialize the parser object, replace the description
    parser = argparse.ArgumentParser(
        description='Map unmapped reads as chimeric fragments',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        'genome_fasta',
        help='Name of genome fasta file. The file must be indexed using'
        'bwa index command prior to this run.')
    parser.add_argument(
        'bamfiles', nargs='+', action='append',
        help='One or more bam files.')
    parser.add_argument(
        '-r', '--reverse_complement', default=False,
        action='store_true',
        help='Treat the reads as reverse complement. This means that the first'
        " read is actually the 3' end of the fragment. Use this when using "
        "Jonathan Livny's protocol for library construction")
    parser.add_argument(
        '-t', '--transcripts',
        help='A gff file of transcripts. If given, screen reads that might'
        ' reside from the same transcript. Very useful for screening ribosomal'
        ' RNAs. Otherwise use only the size limit.')
    parser.add_argument(
        '-s', '--distance', type=int, default=1000,
        help='Maximal distance between concordant reads. If they are generated'
        ' from the same strand but larger than this distance they will be'
        ' considered as chimeric.')
    parser.add_argument(
        '--dust_thr', type=float, default=10,
        help='Threshold for dust filter. If 0 skip.')
    parser.add_argument(
            '-d', '--dirout', default='./remapped-data/',
        help='Output directory, default is this directory.')
    parser.add_argument(
        '-a', '--all_reads',
        help='Map all reads in the BAM file, write all the fragments that are'
        ' not chimeric to the file specified here e.g. '
        '-a single_fragments_mapping.txt. By default these reads will be '
        'written to the standard output.')
    parser.add_argument(
        '-A', '--add_all_reads', default=True, action='store_false',
        help='By default map all reads in the BAM file, write all the fragments'
        ', either chimeric ro single to the output file (stdout). '
        "If this option is selected don't wirte the single reads.")
    parser.add_argument(
        '--keep_circular', default=False, action='store_true',
        help='Remove reads that are probably a result of circular RNAs by'
        ' default. If the reads are close but in opposite order they will be'
        ' removed unless this argument is set.')
    parser.add_argument(
        '-l', '--length', type=int, default=25,
        help='Length of sequence to map. Take the ends of the fragment and map'
        ' each to the genome. The length of the region will be this length.')
    parser.add_argument(
        '--max_mismatches', type=int, default=3,
        help='Find alignment allowing this number of mismatches. If there are '
        'more than one match with this number of mismatches the read will be'
        ' treated as if it might match all of them and if there is one '
        'scenario in which the two ends are concordant it will be removed.')
    parser.add_argument(
        '--allowed_mismatches', type=int, default=1,
        help='This number of mismatches is allowed between the a match and '
        'the genome. If there are mapped reads with less than --max_mismatches'
        ' mismatches but more than this number the read will be ignored.')
    parser.add_argument(
        '--skip_mapping', action='store_true', default=False,
        help='Skip the mapping step, use previously mapped files.')
    parser.add_argument(
        '--maxG', type=float, default=0.8,
        help='If a read has more than this fraction of Gs remove this read'
        'from the screen. This is due to nextseq technology which puts G '
        'where there is no signal, the poly G might just be noise.'
        ' When using other sequencing technologies set to 1.')
    parser.add_argument(
        '-f', '--feature', default='exon',
        help='Name of features to count on the GTF file (column 2).')
    parser.add_argument(
        '-i', '--identifier', default='gene_id',
        help='Name of identifier to print (in column 8 of the GTF file).')
    parser.add_argument(
        '--bwa_exec', default='bwa',
        help='bwa command')
    parser.add_argument(
        '-S', '--samtools_cmd', default='samtools',
        help='Samtools executable.')
    parser.add_argument(
            '--params_aln', default='-t 8 -N -M 0',
        help='Additional parameters for aln function of bwa.')
    parser.add_argument(
        '--samse_params', default='-n 1000',
        help='Additional parameters for samse function of bwa.')
    settings = parser.parse_args(argv)

    return settings


def main(argv=None):
    sys.stderr.write("RILseq version: {}\n".format(pkg_resources.get_distribution("RILseq").version))
    settings = process_command_line(argv)
    # Read the transcripts if given
    try:
        os.makedirs(settings.dirout)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
    if settings.transcripts:
        trans_dict = RILseq.read_transcripts(settings.transcripts, settings.feature, settings.identifier)
    else:
        trans_dict = None
    # Get the ends of the reads from the bam files
#    sys.stderr.write('%s\n'%str(settings.bamfiles))
    if settings.all_reads:
        try:
            outall = open(settings.all_reads, 'w')
        except IOError:
            outall = None
    elif settings.add_all_reads:
        outall = sys.stdout
    else:
        outall = None
    for bf in RILseq.flat_list(settings.bamfiles):
        bfin = pysam.AlignmentFile(bf,'rb')
        outhead = bf.rsplit('.', 1)[0]
        libname = outhead.rsplit('/',1)[-1]
        fsq1name = "%s/%s_ends_1.fastq"%(settings.dirout, libname)
        fsq2name = "%s/%s_ends_2.fastq"%(settings.dirout, libname)
        if settings.skip_mapping:
            fsq1 = open(os.devnull, 'w')
            fsq2 = fsq1
        else:
            fsq1 = open(fsq1name, 'w')
            fsq2 = open(fsq2name, 'w')
        single_mapped = RILseq.get_unmapped_reads(
            bfin, fsq1, fsq2, settings.length, settings.maxG,
            rev=settings.reverse_complement, all_reads=True,
            dust_thr=settings.dust_thr)
        reads_in = []
        # Map the fastq files to the genome
        for fqname in (fsq1name, fsq2name):
            bamheadname = fqname.rsplit('.',1)[0].rsplit('/',1)[-1]
            if settings.skip_mapping:
                bamname = "%s/%s.bam"%(settings.dirout, bamheadname)
            else:
                bamname = RILseq.run_bwa(
                    settings.bwa_exec, fqname, None,
                    os.path.abspath(settings.dirout), bamheadname, settings.max_mismatches,
                    os.path.abspath(settings.genome_fasta), settings.params_aln,
                    '', settings.samse_params,
                    settings.samtools_cmd)
            bamin = pysam.AlignmentFile(bamname,'rb')
            reads_in.append(RILseq.read_bam_file(
                    bamin, bamin.references, settings.allowed_mismatches))
        RILseq.write_reads_table(
            sys.stdout, reads_in[0], reads_in[1], bfin.references,
            settings.distance, not settings.keep_circular,
            trans_dict, write_single=outall, single_mapped=single_mapped,
            max_NM=settings.allowed_mismatches)
    return 0        # success

if __name__ == '__main__':
    status = main()
    sys.exit(status)



#!/usr/bin/env python

"""
This script is used to map fastq files to the genome. The input is a comma
separated list of fastq[.gz] files (or two lists if the input is paired-end).
The output are bam files with the mapped reads, a table containing the number
of reads mapped to each gene and a wiggle file with the coverage of the
reads.
"""

import sys
import argparse
import pysam
import csv
import os
from Bio import SeqIO

import RILseq

def process_command_line(argv):
    """
    Return a 2-tuple: (settings object, args list).
    `argv` is a list of arguments, or `None` for ``sys.argv[1:]``.
    """
    if argv is None:
        argv = sys.argv[1:]

    # initialize the parser object, replace the description
    parser = argparse.ArgumentParser(
        description='Map fastq files to the genome using bwa.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        'genome_fasta',
        help='Name of genome fasta file. The file must be indexed using'
        'bwa index command prior to this run.')
    parser.add_argument(
        '-1', '--fastq_1', action='append', nargs='+',
        help='A list of the first read of the sequencing.')
    parser.add_argument(
        '-2', '--fastq_2', action='append', nargs='*',
        help='A list of the second read of the sequencing.'
        ' The order of these files should be as same as -1. Optional.')
    parser.add_argument(
        '-g', '--genes_gff',
        help='Name of gff file to count the reads per gene. If not given '
        ' or not readable, skip this stage.')
    parser.add_argument(
        '-r', '--reverse_complement', default=False,
        action='store_true',
        help='Treat the reads as reverse complement only when counting'
        ' number of reads per gene and generating wig file. The resulting BAM'
        ' files will be the original ones. Use this when treating libraries'
        " built using Livny's protocol.")
    parser.add_argument(
        '-f', '--feature', default='exon',
        help='Name of features to count on the GTF file (column 2).')
    parser.add_argument(
        '-i', '--identifier', default='gene_id',
        help='Name of identifier to print (in column 8 of the GTF file).')
    parser.add_argument(
        '-v', '--overlap', type=int, default=5,
        help='Minimal required overlap between the fragment and the feature.')
    parser.add_argument(
        '-m', '--allowed_mismatches', type=float, default=2,
        help="Allowed mismatches for BWA mapping.")
    parser.add_argument(
        '-o', '--outhead', default='bwa_mapped_single_reads',
        help='Output file names of counts table (suffixed _counts.txt) and'
        ' wiggle file (suffixed _coverage.wig)')
    parser.add_argument(
        '-d', '--dirout', default='.',
        help='Output directory, default is this directory.')
    parser.add_argument(
        '--bwa_exec', default='bwa',
        help='bwa command')
    parser.add_argument(
        '-S', '--samtools_cmd', default='samtools',
        help='Samtools executable.')
    parser.add_argument(
        '-a', '--params_aln', default='-t 8 -R 200',
        help='Additional parameters for aln function of bwa.')
    parser.add_argument(
        '-s', '--sampe_params', default='-a 1500 -P',
        help='Additional parameters for sampe function of bwa.')
    parser.add_argument(
        '--samse_params', default=' ',
        help='Additional parameters for samse function of bwa.')
    parser.add_argument(
        '-w', '--create_wig', default=False, action='store_true',
        help='Create a coverage wiggle file.')

    settings = parser.parse_args(argv)

    return settings

def main(argv=None):
    settings = process_command_line(argv)
    if not os.path.exists(settings.dirout):
        os.makedirs(settings.dirout)

    genome_len = {}
    if settings.genome_fasta:
        for chrf in SeqIO.parse(settings.genome_fasta, 'fasta'):
            genome_len[chrf.id] = len(chrf.seq)

    if settings.genes_gff:
        try:
            pos_feat_list, all_features = RILseq.read_gtf(
                open(settings.genes_gff), settings.feature, settings.identifier)
        except IOError:
            settings.genes_gff = None
        gcounts = {}
        lib_order = []
    
    fastq_1_list = list(RILseq.flat_list(settings.fastq_1))
    fastq_2_list = list(RILseq.flat_list(settings.fastq_2))
    for i, r1_name in enumerate(RILseq.flat_list(settings.fastq_1)):
        try:
            r2_name = fastq_2_list[i]
        except IndexError:
            r2_name = None
        outhead = r1_name.rsplit('.', 1)[0]
        libname = outhead.rsplit('/',1)[-1]
        outhead = '%s_bwa'%libname
        bamname = RILseq.run_bwa(
            settings.bwa_exec, r1_name, r2_name,
            os.path.abspath(settings.dirout), outhead, settings.allowed_mismatches,
            os.path.abspath(settings.genome_fasta), settings.params_aln, settings.sampe_params,
            settings.samse_params, settings.samtools_cmd)
        samfile = pysam.AlignmentFile(bamname,'rb')
        if settings.genes_gff:
            lib_order.append(libname)
            gcounts[libname] = RILseq.count_features(
                pos_feat_list, samfile, settings.overlap,
                rev=settings.reverse_complement)
        if settings.create_wig:
            outwigs = [open("%s/%s_coverage.wig"%(settings.dirout, fastq.split("_cutadapt")[0].split('/')[-1]), 'w')
               for fastq in fastq_1_list]
            coverage = RILseq.generate_wig(
                samfile, rev=settings.reverse_complement, genome_lengths=genome_len)
            RILseq.print_wiggle(
                coverage, "%s_single_fragments_coverage"%libname,
                "%s single fragments coverage"%libname, outwigs[i])
    # Print the table of counts
    if settings.genes_gff:
        outtables = [open("%s/%s_counts.txt"%(settings.dirout, fastq.split("_cutadapt")[0].split('/')[-1]), 'w')
                     for fastq in fastq_1_list]
        for i, r1_name in enumerate(fastq_1_list):
            outt = csv.writer(outtables[i], delimiter='\t')
            outt.writerow(['Gene name'] + lib_order)
            for g in sorted(list(all_features)):
                row_out = [g]
                for libn in lib_order:
                    row_out.append(gcounts[libn][g])
                outt.writerow(row_out)
    return 0        # success

if __name__ == '__main__':
    status = main()
    sys.exit(status)
